#!/usr/bin/env python

from msvcrt import getwch  # type: ignore

__all__ = ["get_char", "KEYS_MAPPING"]

NULL = "\0"
WIN_ESC = "\xe0"
KEYS_MAPPING = {
    "\xe0\x47": "home",
    "\xe0\x48": "up",
    "\xe0\x49": "pgup",
    "\xe0\x4b": "left",
    "\xe0\x4d": "right",
    "\xe0\x4f": "end",
    "\xe0\x50": "down",
    "\xe0\x51": "pgdn",
    "\xe0\x52": "insert",
    "\xe0\x53": "del",
    "\0\x3b": "f1",
    "\0\x3c": "f2",
    "\0\x3d": "f3",
    "\0\x3e": "f4",
    "\0\x3f": "f5",
    "\0\x40": "f6",
    "\0\x41": "f7",
    "\0\x42": "f8",
    "\0\x43": "f9",
    "\0\x44": "f10",
    "\0\x85": "f11",
    "\0\x86": "f12",
}


def get_char() -> str:
    """
    Read a keypress and return the resulting character as a string.

    Returns:
        char: the pressed key or the key description (e.g. "home")
    """
    ch: str = getwch()
    if ch == WIN_ESC or ch == NULL:  # When reading arrow/insert/del key, the first call returnx 0xe0
        keys_mapping = KEYS_MAPPING
        while keys_mapping:
            ch = ch + getwch()
            keys_mapping = {k: v for k, v in keys_mapping.items() if k.startswith(ch)}
            if len(keys_mapping) == 1 and next(iter(keys_mapping.keys())) == ch:
                result = next(iter(keys_mapping.values()))
                return result
        return ""
    return ch



#!/usr/bin/env python

from .finder import CustomAction, Finder, DEFAULT_POINTER, DEFAULT_PROMPT, DEFAULT_HEADER, DEFAULT_LAYOUT, DEFAULT_MATCHER
from .keys import KeysBinding
from .matcher import Matcher
from .layout import Layout
from .info import InfoStyle
from typing import Any, Callable, Iterator, Optional, Sequence, Type, Union

__version__ = "0.0.13"
"PZP Version"

__all__ = ["pzp", "CustomAction"]


def pzp(
    candidates: Union[Callable[[], Sequence[Any]], Iterator[Any], Sequence[Any]],
    height: Optional[int] = None,
    fullscreen: bool = True,
    format_fn: Callable[[Any], str] = lambda x: str(x),
    layout: Union[Type[Layout], str] = DEFAULT_LAYOUT,
    info_style: InfoStyle = InfoStyle.DEFAULT,
    pointer_str: str = DEFAULT_POINTER,
    prompt_str: str = DEFAULT_PROMPT,
    header_str: str = DEFAULT_HEADER,
    keys_binding: Optional[KeysBinding] = None,
    matcher: Union[Matcher, Type[Matcher], str] = DEFAULT_MATCHER,
    input: Optional[str] = None,
) -> Any:
    """
    Open pzp and return the selected element

    Examples:
        >>> pzp(candidates=list(Path('.').iterdir()))
        PosixPath('README.md')

    Args:
        candidates: Candidates
        fullscreen: Full screen mode
        height: Finder window height
        format_fn: Items format function
        layout: Finder layout
        info_style: Determines the display style of finder info
        pointer_str: Pointer to the current line
        prompt_str: Input prompt
        header_str: Header
        keys_binding: Custom keys binding
        matcher: Matcher

    Returns:
        item: the selected item
    """
    finder = Finder(
        candidates=candidates,
        fullscreen=fullscreen,
        height=height,
        format_fn=format_fn,
        layout=layout,
        info_style=info_style,
        pointer_str=pointer_str,
        prompt_str=prompt_str,
        header_str=header_str,
        keys_binding=keys_binding,
        matcher=matcher,
    )
    return finder.show(input=input)



from pzp.finder import Finder


def test_list():
    candidates = [f"0x{x}" for x in range(0, 100)]
    finder = Finder(candidates=candidates, height=25)
    finder.setup()
    assert finder.layout.screen_items_len == finder.layout.screen.height - 2
    assert finder.layout.config.info_lines == 1
    assert finder.candidates.candidates == candidates
    finder.refresh_candidates()
    assert finder.candidates.candidates == candidates


def test_range():
    candidates = range(0, 100)
    finder = Finder(candidates=candidates, height=25)
    finder.setup()
    assert finder.layout.screen_items_len == finder.layout.screen.height - 2
    assert finder.candidates.candidates == candidates
    finder.refresh_candidates()
    assert finder.candidates.candidates == candidates


def test_yield():
    def get_data():
        yield "a"
        yield "b"
        yield "c"

    finder = Finder(candidates=get_data(), height=25, fullscreen=True)
    finder.setup()
    assert finder.layout.screen_items_len == 3
    assert finder.candidates.candidates == ["a", "b", "c"]


def test_func():
    def get_data():
        return [1, 2, 3]

    finder = Finder(candidates=get_data, height=25, fullscreen=True)
    finder.setup()
    assert finder.layout.screen_items_len == 3
    assert finder.candidates.candidates == [1, 2, 3]
    finder.refresh_candidates()
    assert finder.layout.screen_items_len == 3
    assert finder.candidates.candidates == [1, 2, 3]


def test_func_yield():
    def get_data():
        yield "a"
        yield "b"
        yield "c"

    finder = Finder(candidates=get_data, height=25, fullscreen=True)
    finder.setup()
    assert finder.layout.screen_items_len == 3
    assert finder.candidates.candidates == ["a", "b", "c"]


def test_height():
    data = list(range(0, 10))
    height = 10
    finder = Finder(candidates=data, height=height, fullscreen=False)
    finder.setup()
    assert finder.layout.screen.height == height
    assert finder.layout.screen_items_len == finder.layout.screen.height - 2



from enum import Enum

__all__ = ["InfoStyle"]


class InfoStyle(Enum):
    "Display style of finder info"

    DEFAULT = "default"
    " Display on the next line to the prompt "
    HIDDEN = "hidden"
    " Do not display finder info"



#!/usr/bin/env python

from typing import Optional, TYPE_CHECKING
from .ansi import SPACE
from .actions import Action, ActionsHandler
from .keys import KeyEvent, KeysHandler

if TYPE_CHECKING:
    from .screen import Screen

__all__ = ["LineEditor"]


class LineEditor(ActionsHandler):
    def __init__(self, line: Optional[str] = None, keys_handler: Optional[KeysHandler] = None) -> None:
        """
        Line editor

        Args:
            line: Initial value
            keys_handler: Keys handler
        """
        super().__init__(keys_handler=keys_handler)
        self.line = line or ""
        self.cursor_pos: int = len(self.line)

    def set_cursor_pos(self, cursor_pos: int) -> None:
        "Set cursor position (absolute)"
        self.cursor_pos = max(min(cursor_pos, len(self.line)), 0)

    def adj_cursor_pos(self, characters: int) -> None:
        "Set cursor position (relative to current position)"
        self.cursor_pos = max(min(self.cursor_pos + characters, len(self.line)), 0)

    def insert(self, ch: str) -> None:
        "Insert characters at the current cursor position"
        if ch >= SPACE:  # Add the character to line
            self.line = self.line[0 : self.cursor_pos] + ch + self.line[self.cursor_pos :]
            self.adj_cursor_pos(len(ch))

    @Action("backward-char", keys=["ctrl-b", "left"])
    def backward_char(self) -> None:
        "Move the cursor back one character"
        self.set_cursor_pos(self.cursor_pos - 1)

    @Action("forward-char", keys=["ctrl-f", "right"])
    def forward_char(self) -> None:
        "Move the cursor forward one character"
        self.set_cursor_pos(self.cursor_pos + 1)

    @Action("beginning-of-line", keys=["ctrl-a", "home"])
    def beginning_of_line(self) -> None:
        "Move the cursor to the line start"
        self.set_cursor_pos(0)

    @Action("end-of-line", keys=["ctrl-e", "end"])
    def end_of_line(self) -> None:
        "Move the cursor to the line end"
        self.set_cursor_pos(len(self))

    @Action("backward-delete-char", keys=["ctrl-h", "bspace"])
    def delete_backward_char(self) -> None:
        "Delete the previous character"
        if self.cursor_pos > 0:
            self.line = self.line[: self.cursor_pos - 1] + self.line[self.cursor_pos :]
        self.adj_cursor_pos(-1)

    @Action("delete-char", keys=["ctrl-d", "del"])
    def delete_char(self) -> None:
        "Delete the current character"
        self.line = self.line[: self.cursor_pos] + self.line[self.cursor_pos + 1 :]
        self.adj_cursor_pos(0)

    @Action("default")
    def default(self, key_event: KeyEvent) -> None:
        self.insert(key_event.ch)

    def print(self, screen: "Screen") -> None:
        """
        Print the line and set the cursor position

        Args:
            screen: Screen instance
        """
        screen.write(self.line).move_left(len(self) - self.cursor_pos)

    def __len__(self) -> int:
        "Return line length"
        return len(self.line)

    def __str__(self) -> str:
        "Return the line"
        return self.line



#!/usr/bin/env python

__all__ = [
    "ESC",
    "NL",
    "SPACE",
    "CURSOR_SAVE_POS",
    "CURSOR_RESTORE_POS",
    "ERASE_LINE",
    "BLACK",
    "RED",
    "GREEN",
    "YELLOW",
    "BLUE",
    "PURPLE",
    "CYAN",
    "WHITE",
    "BLACK_BG",
    "RED_BG",
    "GREEN_BG",
    "YELLOW_BG",
    "BLUE_BG",
    "PURPLE_BG",
    "CYAN_BG",
    "WHITE_BG",
    "RESET",
    "BOLD",
    "NEGATIVE",
]

ESC = "\u001b"
NL = "\n"
SPACE = " "

CURSOR_SAVE_POS = f"{ESC}7"
"Save cursor position"
CURSOR_RESTORE_POS = f"{ESC}8"
"Restores the cursor to the last saved position"
ERASE_LINE = f"{ESC}[2K"
"Erase the entire line"

BLACK = f"{ESC}[30m"
"Black foreground color"
RED = f"{ESC}[31m"
"Red foreground color"
GREEN = f"{ESC}[32m"
"Green foreground color"
YELLOW = f"{ESC}[33m"
"Yellow foreground color"
BLUE = f"{ESC}[34m"
"Blue foreground color"
PURPLE = f"{ESC}[35m"
"Purple foreground color"
CYAN = f"{ESC}[36m"
"Cyan foreground color"
WHITE = f"{ESC}[37m"
"White foreground color"

BLACK_BG = f"{ESC}[40m"
"Black background color"
RED_BG = f"{ESC}[41m"
"Red background color"
GREEN_BG = f"{ESC}[42m"
"Green background color"
YELLOW_BG = f"{ESC}[43m"
"Yellow background color"
BLUE_BG = f"{ESC}[44m"
"Blue background color"
PURPLE_BG = f"{ESC}[45m"
"Purple background color"
CYAN_BG = f"{ESC}[46m"
"Cyan background color"
WHITE_BG = f"{ESC}[47m"
"White background color"

RESET = f"{ESC}[0m"
"Reset styles and colors"
BOLD = f"{ESC}[1m"
"Set bold mode"
NEGATIVE = f"{ESC}[7m"
"Set inverse mode"



#!/usr/bin/env python

import re
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Sequence, Type, Tuple, Union
from .commons import first_or_default

__all__ = [
    "Matcher",
    "ExactMatcher",
    "ExtendedMatcher",
    "get_matcher",
    "list_matchers",
]

SPLIT_ESCAPED_RE = re.compile(r"(?<!\\)\s+")
PREFIXES = ["!^", "!'", "^", "'", "!"]
SUFFIXES = ["$"]

matchers: Dict[str, Type["Matcher"]] = {}


class Matcher(ABC):
    def __init_subclass__(cls, option: str, **kwargs: Dict[str, Any]) -> None:
        "Register a subclass"
        super().__init_subclass__(**kwargs)
        matchers[option] = cls

    @abstractmethod
    def filter(self, pattern: str, candidates: Sequence[Any], format_fn: Callable[[Any], str] = lambda x: str(x)) -> Sequence[Any]:
        """
        Filter candidates according to the given pattern

        Args:
            pattern: Pattern
            candidates: Candidates
            format_fn: Items format function

        Returns:
            result: Filtered candidates
        """
        pass  # pragma: no cover


class ExactMatcher(Matcher, option="exact"):
    """
    Exact matcher
    """

    def filter(self, pattern: str, candidates: Sequence[Any], format_fn: Callable[[Any], str] = lambda x: str(x)) -> Sequence[Any]:
        pattern = pattern.lower()
        return [item for item in candidates if pattern in format_fn(item).lower()]


class ExtendedMatcherFilter:
    def __init__(self, term: str) -> None:
        self.prefix, self.term, self.suffix = self.split(term)
        # If a term is prefixed by !, the filter will exclude the lines that satisfy the term
        if "!" in self.prefix:
            self.exclude = True
        else:
            self.exclude = False
        if not self.term:
            self.exclude = False
            self.fn = self.always
        elif "'" in self.prefix:
            # Quoted - Check if the text contains the given term
            self.fn = self.in_match
        elif "^" in self.prefix:
            # If a term is prefixed by ^, the filter will include the lines than start with the given term
            if "$" in self.suffix:
                self.fn = self.exact_match
            else:
                self.fn = self.startswith_match
        elif "$" in self.suffix:
            # If a term is suffixed by $, the filter will include the lines than end with the given term
            self.fn = self.endswith_match
        else:
            # Check if the text contains the given term
            self.fn = self.in_match

    @classmethod
    def split(cls, term: str) -> Tuple[str, str, str]:
        "Split term in prefix, term, suffix"
        term, prefix = first_or_default([(term[len(p) :], p) for p in PREFIXES if term.startswith(p)], (term, ""))
        term, suffix = first_or_default([(term[: len(term) - len(s)], s) for s in SUFFIXES if term.endswith(s)], (term, ""))
        return prefix, term, suffix

    def __call__(self, txt: str) -> bool:
        "Evaluate the filter on the given text"
        if self.exclude:
            return not self.fn(txt)
        else:
            return self.fn(txt)

    def startswith_match(self, txt: str) -> bool:
        "True if the text startswith the given term"
        return txt.startswith(self.term)

    def endswith_match(self, txt: str) -> bool:
        "True if the text endswith the given term"
        return txt.endswith(self.term)

    def in_match(self, txt: str) -> bool:
        "True if the text contains the given term"
        return self.term in txt

    def exact_match(self, txt: str) -> bool:
        "Exact match"
        return self.term == txt

    def always(self, txt: str) -> bool:
        "Always return True"
        return True

    def __str__(self) -> str:
        return f"<{self.prefix}, {self.term}, {self.suffix}>"  # pragma: no cover


class ExtendedMatcher(Matcher, option="extended"):
    """
    Extended Matcher

    This matcher accept multiple patterns delimited by spaces, such as: term ^start end$ !not

    If patter is prefixed by a single-quote character ', it will not be splitted by spaces.

    A backslash can be prepend to a space to match a literal space character.

    A term can be prefixed by ^, or suffixed by $ to become an anchored-match term.
    Then matcher will search for the lines that start with or end with the given string.

    If a term is prefixed by !, the matcher will exclude the lines that satisfy the term from the result.
    """

    def filter(self, pattern: str, candidates: Sequence[Any], format_fn: Callable[[Any], str] = lambda x: str(x)) -> Sequence[Any]:
        # Prepare the filters
        filters = [ExtendedMatcherFilter(term) for term in self.split_pattern(pattern)]
        # Filter items
        return [item for item in candidates if self.filter_item(filters, item, format_fn)]

    def filter_item(
        self, filters: Sequence[ExtendedMatcherFilter], item: Any, format_fn: Callable[[Any], str] = lambda x: str(x)
    ) -> bool:
        txt: str = format_fn(item).lower()
        return all([filter_fn(txt) for filter_fn in filters])

    def split_pattern(self, pattern: str) -> Sequence[str]:
        "Split a pattern into terms"
        if pattern.startswith("'") or pattern.startswith("!'"):  # quote - keep spaces
            return [pattern.lower()]
        else:
            return [term.replace("\\", "") for term in SPLIT_ESCAPED_RE.split(pattern.lower()) if term]


def get_matcher(matcher: Union[Matcher, Type[Matcher], str]) -> Matcher:
    "Get a matcher instance by name or by class"
    if isinstance(matcher, Matcher):
        return matcher
    elif isinstance(matcher, str):
        return matchers[matcher]()
    else:
        return matcher()


def list_matchers() -> Sequence[str]:
    "List matchers"
    return list(matchers.keys())



from pathlib import Path
from pzp.finder import Finder
from pzp.matcher import ExactMatcher, ExtendedMatcher, ExtendedMatcherFilter

tests_dir = Path(__file__).parent
code_of_conduct = tests_dir.parent / "CODE_OF_CONDUCT.md"


class LineItem:
    def __init__(self, i, line):
        self.i = i
        self.line = line

    def __str__(self):
        return self.line


def test_extended_matcher_filter():
    t = ExtendedMatcherFilter("")
    assert t.prefix == "" and t.suffix == "" and t.term == ""
    t = ExtendedMatcherFilter("aaa")
    assert t.prefix == "" and t.suffix == "" and t.term == "aaa"
    t = ExtendedMatcherFilter("!aaa")
    assert t.prefix == "!" and t.suffix == "" and t.term == "aaa"
    t = ExtendedMatcherFilter("!'aaa")
    assert t.prefix == "!'" and t.suffix == "" and t.term == "aaa"
    t = ExtendedMatcherFilter("^aaa")
    assert t.prefix == "^" and t.suffix == "" and t.term == "aaa"
    t = ExtendedMatcherFilter("!^aaa")
    assert t.prefix == "!^" and t.suffix == "" and t.term == "aaa"
    t = ExtendedMatcherFilter("^aaa$")
    assert t.prefix == "^" and t.suffix == "$" and t.term == "aaa"
    t = ExtendedMatcherFilter("!^aaa$")
    assert t.prefix == "!^" and t.suffix == "$" and t.term == "aaa"
    t = ExtendedMatcherFilter("!aaa$")
    assert t.prefix == "!" and t.suffix == "$" and t.term == "aaa"
    t = ExtendedMatcherFilter("aaa$")
    assert t.prefix == "" and t.suffix == "$" and t.term == "aaa"


def test_matcher_in():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = ""
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(finder.candidates.candidates)

    finder.line_editor.line = "ex"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 13
    assert all(["ex" in item.lower() for item in finder.candidates.matching_candidates])

    finder.line_editor.line = "ex e"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 13
    assert all(["ex" in item.lower() for item in finder.candidates.matching_candidates])

    finder.line_editor.line = "ex ex"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 13
    assert all(["ex" in item.lower() for item in finder.candidates.matching_candidates])


def test_matcher_not_in():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = "!ex"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(candidates) - 13
    assert all(["ex" not in item.lower() for item in finder.candidates.matching_candidates])

    finder.line_editor.line = "!ex ex"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 0

    finder.line_editor.line = "!ex We"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 7
    assert all(["ex" not in item.lower() for item in finder.candidates.matching_candidates])
    assert all(["we" in item.lower() for item in finder.candidates.matching_candidates])

    finder.line_editor.line = "!"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(candidates)


def test_matcher_startswith():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = "^We"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 2

    finder.line_editor.line = "^We as"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 1

    finder.line_editor.line = "as ^We"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 1

    finder.line_editor.line = "as ^bla"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 0


def test_matcher_not_startswith():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = "!^We"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(candidates) - 2

    finder.line_editor.line = "!^We as"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 12

    finder.line_editor.line = "as !^We"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 12

    finder.line_editor.line = "!^bla"
    finder.apply_filter()


def test_matcher_endswith():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = ",$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 6

    finder.line_editor.line = "s,$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 3

    finder.line_editor.line = "* s,$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 1


def test_matcher_not_endswith():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = "!,$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(candidates) - 6

    finder.line_editor.line = "!s,$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(candidates) - 3

    finder.line_editor.line = "* !s,$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 17


def test_quoted():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = "'  "
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 6

    finder.line_editor.line = "'to this"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 1

    finder.line_editor.line = "!'to this"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == len(candidates) - 1


def test_line_exact_match():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True)
    finder.setup()

    finder.line_editor.line = "^##\\ Scope$"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 1


def test_exact_matcher():
    candidates = code_of_conduct.read_text().split("\n")
    finder = Finder(candidates=candidates, height=10, fullscreen=True, matcher=ExactMatcher)
    finder.setup()

    finder.line_editor.line = "  "
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 6

    finder.line_editor.line = "to this"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 1

    finder.line_editor.line = "!'to this"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 0


def test_extended_matcher():
    matcher = ExtendedMatcher()
    candidates = range(0, 1000)
    result = matcher.filter("0", candidates)
    assert len(result) == 181
    result = matcher.filter("9", candidates)
    assert len(result) == 271
    result = matcher.filter("0 1", candidates)
    assert len(result) == 36
    result = matcher.filter("0 !1", candidates)
    assert len(result) == 181 - 36
    result = matcher.filter("!^0", candidates)
    assert len(result) == 999


def test_matcher_format_fn():
    candidates = [LineItem(i, line) for i, line in enumerate(code_of_conduct.read_text().split("\n"), start=1)]
    finder = Finder(
        candidates=candidates,
        height=10,
        fullscreen=True,
        format_fn=lambda item: f"{item.i:3d}: {item.line}",
    )
    finder.setup()

    finder.line_editor.line = "ex"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 13
    assert all(["ex" in item.line.lower() for item in finder.candidates.matching_candidates])

    finder.line_editor.line = "20:"
    finder.apply_filter()
    assert len(finder.candidates.matching_candidates) == 2
    assert all(["20" in str(x.i) for x in finder.candidates.matching_candidates])



#!/usr/bin/env python

from typing import Any

__all__ = [
    "PZPException",
    "GenericAction",
    "AcceptAction",
    "AbortAction",
    "CustomAction",
    "MissingHander",
]


class PZPException(Exception):
    """
    Generic PZP Exception
    """


class GenericAction(PZPException):
    """
    Generic Action Event

    Args:
        action: action
        ch: pressed key
        selected_item: selected item, if any

    Attributes:
        action: action
        ch: pressed key
        selected_item: selected item, if any
    """

    def __init__(self, action: str, ch: str, selected_item: Any = None):
        super().__init__(action)
        self.action = action
        self.ch = ch
        self.selected_item = selected_item


class AcceptAction(GenericAction):
    """
    The AcceptAction exception is raised when the user presses a key
    that is mapped to the "accept" action.

    Args:
        action: action
        ch: pressed key
        selected_item: selected item, if any

    Attributes:
        action: action
        ch: pressed key
        selected_item: selected item, if any
    """

    def __init__(self, action: str, ch: str, selected_item: Any = None):
        super().__init__(action, ch, selected_item)


class AbortAction(GenericAction):
    """
    The AbortAction exception is raised when the user presses a key
    that is mapped to the "abort" action.

    Args:
        action: action
        ch: pressed key
        selected_item: selected item, if any

    Attributes:
        action: action
        ch: pressed key
        selected_item: selected item, if any
    """

    def __init__(self, action: str, ch: str, selected_item: Any = None):
        super().__init__(action, ch, selected_item)


class CustomAction(GenericAction):
    """
    The CustomAction exception is raised when the user presses a key
    that is mapped to the "custom" action.

    Args:
        action: action
        selected_item: selected item, if any
        ch: pressed key

    Attributes:
        action: action
        selected_item: selected item, if any
        ch: pressed key
    """

    def __init__(self, action: str, ch: str, selected_item: Any = None):
        super().__init__(action, ch, selected_item)


class MissingHander(PZPException):
    """
    The MissingHander exception is raised when there is no handler for an action.

    Args:
        action: action
        ch: pressed key

    Attributes:
        action: action
        ch: pressed key
    """

    def __init__(self, action: str, ch: str):
        super().__init__(action)
        self.action = action
        self.ch = ch



#!/usr/bin/env python

import sys
from typing import Any, Callable, Iterator, Optional, Sequence, TextIO, Type, Union
from .actions import Action, ActionsHandler
from .candidates import Candidates
from .config import Config
from .exceptions import AcceptAction, AbortAction, CustomAction, MissingHander
from .info import InfoStyle
from .keys import KeyEvent, KeysBinding
from .layout import Layout, get_layout
from .line_editor import LineEditor
from .matcher import Matcher

__all__ = [
    "Finder",
    "CustomAction",
    "DEFAULT_POINTER",
    "DEFAULT_PROMPT",
    "DEFAULT_HEADER",
    "DEFAULT_LAYOUT",
    "DEFAULT_MATCHER",
]

DEFAULT_POINTER = ">"
"Default pointer"
DEFAULT_PROMPT = "> "
"Default input prompt"
DEFAULT_HEADER = ""
"Default header"
DEFAULT_LAYOUT = "default"
"Default layout"
DEFAULT_MATCHER = "extended"
"Default matcher"


class Finder(ActionsHandler):
    def __init__(
        self,
        candidates: Union[Callable[[], Sequence[Any]], Iterator[Any], Sequence[Any]],
        fullscreen: bool = True,
        height: Optional[int] = None,
        format_fn: Callable[[Any], str] = lambda x: str(x),
        layout: Union[Type[Layout], str] = DEFAULT_LAYOUT,
        info_style: InfoStyle = InfoStyle.DEFAULT,
        pointer_str: str = DEFAULT_POINTER,
        prompt_str: str = DEFAULT_PROMPT,
        header_str: str = DEFAULT_HEADER,
        keys_binding: Optional[KeysBinding] = None,
        matcher: Union[Matcher, Type[Matcher], str] = DEFAULT_MATCHER,
        output_stream: TextIO = sys.stderr,
    ):
        """
        Initializate Finder object

        Args:
            candidates: Candidates
            fullscreen: Full screen mode
            height: Finder window height
            format_fn: Items format function
            layout: Finder layout
            info_style: Determines the display style of finder info
            pointer_str: Pointer to the current line
            prompt_str: Input prompt
            header_str: Header
            keys_binding: Custom keys binding
            matcher: Matcher
            output_stream: Output stream
        """
        super().__init__(keys_binding=keys_binding)
        self.config = Config(fullscreen, height, format_fn, info_style, pointer_str, prompt_str, header_str, output_stream)
        self.candidates = Candidates(candidates=candidates, format_fn=format_fn, matcher=matcher)
        self.layout: Layout = get_layout(layout=layout, config=self.config, candidates=self.candidates)

    def setup(self, input: Optional[str] = None) -> None:
        """
        Setup Finder execution

        Args:
            input: initial search string
        """
        self.line_editor = LineEditor(line=input or "", keys_handler=self.keys_handler)
        # Load the candidate list
        self.refresh_candidates()
        # Filter the items, calculate the screen offset
        self.apply_filter()
        # Calculate the required height and setup the screen
        self.layout.screen_setup(self.line_editor)

    def show(self, input: Optional[str] = None) -> Any:
        """
        Open pzp and return the selected element

        Args:
            input: initial search string

        Returns:
            item: the selected item
        """
        self.setup(input=input)
        try:
            while True:
                self.process_key()
                self.apply_filter()
                self.update_screen()
        except AcceptAction as accept:
            return accept.selected_item
        except AbortAction:
            return None
        finally:
            self.layout.cleanup()

    def refresh_candidates(self) -> None:
        "Load/reload the candidate list"
        self.candidates.refresh_candidates()
        self.selected: int = 0

    def process_key(self, ch: Optional[str] = None) -> None:
        "Process the pressed key"
        key_event = self.keys_handler.get_key_event(ch)
        try:
            self.line_editor.process_key_event(key_event)
        except MissingHander:
            try:
                self.process_key_event(key_event)
            except MissingHander:
                raise CustomAction(action=key_event.action, ch=key_event.ch, selected_item=self.prepare_result())  # type: ignore

    def apply_filter(self) -> None:
        "Filter the items"
        self.candidates.apply_filter(pattern=str(self.line_editor))
        self.selected = max(min(self.selected, self.candidates.matching_candidates_len - 1), 0)

    def update_screen(self) -> None:
        "Update the screen - erase the old items, print the filtered items and the prompt"
        self.layout.update_screen(selected=self.selected)

    def prepare_result(self) -> Any:
        "Output the selected item, if any"
        try:
            return self.candidates.matching_candidates[self.selected]
        except IndexError:
            return None

    @Action("accept", keys=["enter"])
    def accept(self, key_event: KeyEvent) -> None:
        "Confirm"
        raise AcceptAction(action="accept", ch=key_event.ch, selected_item=self.prepare_result())

    @Action("abort", keys=["ctrl-c", "ctrl-g", "ctrl-q", "esc"])
    def abort(self, key_event: KeyEvent) -> None:
        "Cancel"
        raise AbortAction(action="abort", ch=key_event.ch)

    @Action("down", keys=["ctrl-j", "ctrl-n", "down"])
    def down(self) -> None:
        "Move one line down"
        self.selected = self.layout.move_selection(self.selected, lines=-1)

    @Action("up", keys=["ctrl-k", "ctrl-p", "up"])
    def up(self) -> None:
        "Move one line up"
        self.selected = self.layout.move_selection(self.selected, lines=+1)

    @Action("page-down", keys=["page-down", "pgdn"])
    def page_down(self) -> None:
        "Move one page down"
        self.selected = self.layout.move_selection(self.selected, pages=-1)

    @Action("page-up", keys=["page-up", "pgup"])
    def page_up(self) -> None:
        "Move one page up"
        self.selected = self.layout.move_selection(self.selected, pages=+1)

    @Action("ignore", keys=["null", "insert"])
    def ignore(self) -> None:
        "Do nothing"
        pass



from setuptools import setup, find_packages
from pzp import __version__


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='pzp',
    version=__version__,
    description="fzf Python clone",
    long_description="fzf Python clone",
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    keywords='struct',
    author='Andrea Bonomi',
    author_email='andrea.bonomi@gmail.com',
    url='http://github.com/andreax79/pzp',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'examples']),
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        # -*- Extra requirements: -*-
    ],
    entry_points="""
      # -*- Entry points: -*-
      """,
    test_suite='tests',
    tests_require=['pytest'],
)



#!/usr/bin/env python

__all__ = ["get_char", "KEYS_MAPPING"]

import sys
import termios
import tty

NULL = "\0"
ESC = "\x1b"
KEYS_MAPPING = {
    "[A": "up",
    "[B": "down",
    "[C": "right",
    "[D": "left",
    "[Z": "btab",  # shift-tab
    "[1;2A": "shift-up",
    "[1;2B": "shift-down",
    "[1;2C": "shift-right",
    "[1;2D": "shift-left",
    "[1~": "home",
    "[2~": "insert",
    "[3~": "del",
    "[4~": "end",
    "[5~": "pgup",
    "[6~": "pgdn",
    "OP": "f1",
    "[11~": "f1",
    "OQ": "f2",
    "[12~": "f2",
    "OR": "f3",
    "[13~": "f3",
    "OS": "f4",
    "[14~": "f4",
    "[15~": "f5",
    "[17~": "f6",
    "[18~": "f7",
    "[19~": "f8",
    "[20~": "f9",
    "[21~": "f10",
    "[23~": "f11",
    "[24~": "f12",
}


def get_char() -> str:
    """
    Read a keypress and return the resulting character as a string.

    Returns:
        char: the pressed key or the key description (e.g. "home")
    """
    try:
        fd = sys.stdin.fileno()
        attrs = termios.tcgetattr(fd)
        tty.setraw(fd)
        ch = sys.stdin.read(1)
        if ch == ESC:
            keys_mapping = KEYS_MAPPING
            ch = ""
            while keys_mapping:
                ch = ch + sys.stdin.read(1)
                keys_mapping = {k: v for k, v in keys_mapping.items() if k.startswith(ch)}
                if len(keys_mapping) == 1 and next(iter(keys_mapping.keys())) == ch:
                    result = next(iter(keys_mapping.values()))
                    return result
            return ""
        return ch
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, attrs)
